/**********************************************************************
*
* Copyright (c) 2012-2020 Barbara Geller
* Copyright (c) 2012-2020 Ansel Sermersheim
*
* You may use this file under the terms of the 2-Clause BSD license
* provided with PepperMill or available at:
*
* https://opensource.org/licenses/BSD-2-Clause
*
* PepperMill is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*
***********************************************************************/

#include "peppermill.h"

#include <QDir>
#include <QFile>
#include <QFileInfoList>
#include <QVector>

#include <stdexcept>
#include <stdio.h>

PepperMill::PepperMill(enum levels type, QString path, bool forceBU)
{
   m_type    = type;
   m_path    = path;
   m_forceBU = forceBU;
}

PepperMill::~PepperMill()
{
}

void PepperMill::findRecursive(const QString &path)
{
   QStringList fileType;
   fileType.append("*.h");

   QDir dir(path);
   dir.setFilter(QDir::AllDirs | QDir::Files | QDir::NoDotAndDotDot | QDir::NoSymLinks);

   QFileInfoList list = dir.entryInfoList(fileType);

   for (int k = 0; k < list.count(); k++) {

      QString filePath = list[k].filePath();

      if ( list[k].isDir() )  {
         // recursive
         findRecursive(filePath);

      } else  {
         m_fileList.append(filePath);

      }
   }
}

void PepperMill::process()
{
   QStringList fileType;
   fileType.append("*.h");

   //
   QDir dir(m_path);
   dir.setFilter(QDir::Files | QDir::NoDotAndDotDot | QDir::NoSymLinks);

   switch (m_type) {

      case ONEFILE:
         // --onefile z:\peppermill\peppermill.h -f

         {
            QFile fileX(m_path);

            if (! fileX.exists()) {
               printf("%s%s\n", "Requested file name does not exist:  ", qPrintable(m_path));
               throw std::runtime_error("");
            }

            m_fileList.append(fileX.fileName());
         }

         break;

      case ONEFOLDER:
         // --onefolder test_Files -f

         if (! dir.exists()) {
            printf("%s%s\n", "Requested folder does not exist:  ", qPrintable(m_path));
            throw std::runtime_error("");
         }

         {
            QFileInfoList list = dir.entryInfoList(fileType);

            for (int k = 0; k < list.count(); k++) {
               m_fileList.append(list[k].filePath());
            }
         }

         break;

      case RECURSIVE:
         // --recursive "z:\peppermill\"

         if (! dir.exists()) {
            printf("%s%s\n", "Requested path does not exist:  ", qPrintable(m_path));
            throw std::runtime_error("");
         }

         findRecursive(m_path);
         break;
   }

   m_sectionSet.insert("public");
   m_sectionSet.insert("protected");
   m_sectionSet.insert("private");
   m_sectionSet.insert("signals");
   m_sectionSet.insert("Q_SIGNALS");

   //
   QString name;
   int processed = 0;

   printf("\n\n%s\n\n", "** Processing, please wait...");

   for (int k = 0; k < m_fileList.size(); ++k) {

      try {
         name = m_fileList.at(k);

         printf("%s%s\n", "   Processing file:  ", qPrintable(name) );
         this->pmConvert(name);
         ++processed;

      } catch (std::exception &e) {
         // move to next file
         printf("%s%s \n\n", "   ", e.what() );

      }
   }

   printf("\n\n%s\n", "Processing Finished");
   printf("\n%s%d\n%s%d\n\n", "# of Files Found: ", m_fileList.size(), "# of Files Processed: ", processed );

}

//
void PepperMill::pmConvert(const QString &name)
{
   bool retval = false;

   QString backUpName = name.left( name.size()-2 );
   backUpName = backUpName + ".cs";

   if (QFile::exists(backUpName))  {

      if (m_forceBU) {
         retval = QFile::remove(backUpName);

         if (retval)  {
            retval = QFile::rename(name, backUpName);
         }

      } else {
         QString errMsg = QString("Back Up file name already exists:  %1").formatArg(backUpName);
         throw std::runtime_error(errMsg.toStdString());

      }

   } else {
      retval = QFile::rename(name, backUpName);

   }

   if (retval)  {

      QFile file(backUpName);

      if (! file.open(QFile::ReadOnly)) {
         QString errMsg = QString("Unable to open or read file:  %1, %2").formatArg(name).formatArg(file.errorString());
         throw std::runtime_error(errMsg.toStdString());
      }

      QByteArray input = file.readAll();

      if (input.isEmpty()) {
         // no need to process this file
         file.close();

      } else {

         try  {
            input = this->pmParse(input);
            file.close();

            file.setFileName(name);

            if (! file.open(QFile::WriteOnly | QFile::Truncate)) {
               QString errMsg = QString("Unable to open or write file:  %1, %2").formatArg(name).formatArg(file.errorString());
               throw std::runtime_error(errMsg.toStdString());
            }

            // save new file
            file.write(input);
            file.close();

         } catch (std::exception &e) {
            QString errMsg = QString("Unable to process file:  %1, %2").formatArg(name).formatArg(QString::fromUtf8(e.what()));
            throw std::runtime_error(errMsg.toStdString());
         }

      }

   } else  {
      // unable to back up file
      throw std::runtime_error("Rename file failed");
   }

}
