/**********************************************************************
*
* Copyright (c) 2012-2020 Barbara Geller
* Copyright (c) 2012-2020 Ansel Sermersheim
*
* You may use this file under the terms of the 2-Clause BSD license
* provided with PepperMill or available at:
*
* https://opensource.org/licenses/BSD-2-Clause
*
* PepperMill is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*
***********************************************************************/

#include "peppermill.h"

#include <stdexcept>

QString PepperMill::pmGetData(const char **ptr)
{
   QString data;

   // a
   while (**ptr) {
      char letter = **ptr;

      if (letter == ' ') {
         // throw out

      } else if (letter == '(') {
         // throw out letter
         ++(*ptr);
         break;

      }

      ++(*ptr);
   }

   // b
   int nested = 0;

   while (**ptr) {
      char letter = **ptr;

      if (letter == '(') {
         ++nested;

      } else if (letter == ')') {

         if (nested == 0) {
            // throw out letter
            ++(*ptr);
            break;

         } else {
            --nested;

         }
      }

      data += letter;
      ++(*ptr);
   }

   return data;
}

QString PepperMill::pmGetMethod(const char **ptr)
{
   QString data;

   int parenLevel = 0;
   int curlyLevel = 0;

   while (true) {
      char letter = **ptr;

      //
      if (letter == '(')  {
         ++parenLevel;

      } else if (letter == ')')  {
         --parenLevel;

      } else if (letter == '{')  {
         ++curlyLevel;

      } else if (letter == '}')  {
         --curlyLevel;

         if (parenLevel == 0 && curlyLevel == 0) {
            // all done
            data += letter;
            break;
         }

      } else if (letter == ';' && curlyLevel == 0)   {
         // all done
         break;
      }

      data += letter;
      ++(*ptr);
   }

   return data;
}

std::tuple<QString, QList<QString>, QString> PepperMill::pmGetSignature(const QString &fullName)
{
   QString methodName;
   QString signature;
   QList<QString> paramNames;

   QStringList tokens;
   QString item;

   // part 1, decipher tokens from fullName
   for (auto letter : fullName) {

      if (letter.isSpace()) {

         // store the token
         if (! item.isEmpty()) {
            tokens.append(item);
            item.clear();
         }

      } else if (letter == '*' || letter == '&' || letter == '=' || letter == ',' ||
                 letter == '[' || letter == ']' ||
                 letter == '(' || letter == ')' ||
                 letter == '{' || letter == '}' ||
                 letter == '<' || letter == '>'  )   {

         // store the token
         if (! item.isEmpty()) {
            tokens.append(item);
            item.clear();
         }

         tokens.append( QString(letter) );

      } else {
         // tack letter onto the current item
         item = item + letter;

      }
   }

   if (! item.isEmpty()) {
      tokens.append(item);
   }

   // part 2, parse return type from tokens
   int index = 0;
   QString typeReturn;

   QString word;
   QString nextWord;

   bool isStar = false;

   int angleLevel   = 0;
   int bracketLevel = 0;
   int parenLevel   = 0;

   while (true)  {
      word = tokens[index];

      // A
      if (word == "<") {
         ++angleLevel;

      } else if (word == ">")  {
         --angleLevel;

      } else if (word == "[")  {
         ++bracketLevel;

      } else if (word == "]")  {
         --bracketLevel;

      } else if (word == "(")  {
         ++parenLevel;

      } else if (word == ")")  {
         --parenLevel;

      }


      // B
      if (word == "const")  {
         nextWord = tokens[index+1];

         if (isStar && nextWord != "*") {
            typeReturn += word;
            ++index;
            break;
         }

      } else if (word == "static")  {
         typeReturn = typeReturn + word + " ";
         ++index;
         continue;

      } else if (word == "inline")  {
         ++index;
         continue;

      } else if (word == "virtual")  {
         typeReturn = typeReturn + word + " ";
         ++index;
         continue;

      } else if (word == "volatile")  {
         typeReturn = typeReturn + word + " ";
         ++index;
         continue;

      } else if (word == "*") {
         isStar   = true;
         nextWord = tokens[index+1];

         if (nextWord == "*" || nextWord == "const") {
            // do nothing

         } else {
            typeReturn += word;
            ++index;
            break;
         }

      } else if (angleLevel != 0 && word == ">")   {
         typeReturn += word;
         ++index;
         continue;

      } else if (angleLevel != 0 && word == "<")   {
         typeReturn += word;
         ++index;
         continue;

      } else if (bracketLevel != 0 && word == "]") {
         typeReturn += word;
         ++index;
         continue;

      } else if (bracketLevel != 0 && word == "[") {
         typeReturn += word;
         ++index;
         continue;

      } else if (parenLevel != 0 && word == ",") {
         isStar = false;

         typeReturn += word;
         ++index;
         continue;

      } else if (parenLevel != 0 && word == ")") {
         isStar = false;

         typeReturn += word;
         ++index;
         continue;

      } else if (parenLevel != 0 && word == "(") {
         typeReturn += word;
         ++index;
         continue;

      } else {
         nextWord = tokens[index+1];

         if (nextWord == "*" || nextWord == "&" || nextWord == "<" || nextWord == "[" || nextWord == "(") {
            typeReturn += word;
            ++index;
            continue;

         }  else {
            typeReturn += word;
            ++index;

            if (angleLevel == 0 && bracketLevel == 0 && parenLevel == 0) {
               break;

            } else {
               continue;

            }
         }
      }

      typeReturn = typeReturn + word + " ";
      ++index;
   }

   // convert return type to "char *" data type
   // unused      returnType = strdup(qPrintable(typeReturn));

   // part 3 parse method name
   methodName = tokens[index++];

   signature  = typeReturn + " " + methodName;

   // part 4, parse parameter list from tokens
   QString typeArg;

   bool found_funcPtrVar;
   bool isDefaultArg;

   int curlyLevel;

   // parse open paren
   if (tokens[index++] == "(")  {
      signature += "(";

   } else  {
      // failed to find a valid methodName
      QString errMsg = "\n\n       Exception: getSignature(), failed to find a valid methodName\n       Parsing: ";
      errMsg += fullName;

      throw std::runtime_error(errMsg.toStdString());
   }

   for (int k = index; k < tokens.size(); ++k) {
      word = tokens[k];

      if (word == ")")  {
         signature += word;
         break;

      } else {
         found_funcPtrVar = false;

         isDefaultArg  = false;
         isStar        = false;
         typeArg       = "";

         angleLevel    = 0;
         bracketLevel  = 0;
         parenLevel    = 0;
         curlyLevel    = 0;

         while (true)  {
            word = tokens[k];

            // A
            if (word == "<") {
               ++angleLevel;

            } else if (word == ">")  {
               --angleLevel;

            } else if (word == "[")  {
               ++bracketLevel;

            } else if (word == "]")  {
               --bracketLevel;

            } else if (word == "(")  {
               ++parenLevel;

            } else if (word == ")")  {
               --parenLevel;

            } else if (word == "{")  {
               ++curlyLevel;

            } else if (word == "}")  {
               --curlyLevel;

            }

            // B
            if (word == "const")  {
               nextWord = tokens[k+1];

               if (isStar && nextWord != "*") {
                  typeArg += word;
                  ++k;
                  break;
               }

            } else if (word == "restrict")  {
               // discard
               ++k;
               continue;

            } else if (word == "volatile")  {
               typeArg = typeArg + word + " ";
               ++k;
               continue;

            } else if (word == "*") {
               isStar   = true;
               nextWord = tokens[k+1];

               if (nextWord == "*" || nextWord == "const") {
                  typeArg += word;
                  ++k;
                  continue;

               }  else if (parenLevel != 0 && nextWord == ")")  {
                  // function pointer with no variable name
                  typeArg += word;
                  ++k;
                  continue;

               }  else if (parenLevel != 0 && ! found_funcPtrVar)  {
                  // function pointer, save the var name, then keep going
                  found_funcPtrVar = true;
                  paramNames.append(nextWord);

                  typeArg += word;
                  k += 2;
                  continue;

               } else  {
                  typeArg += word;
                  ++k;

                  if (parenLevel == 0) {
                     break;
                  } else {
                     continue;
                  }

               }

            } else if (angleLevel != 0 && word == ">")   {
               typeArg += word;
               ++k;
               continue;

            } else if (angleLevel != 0 && word == "<")   {
               typeArg += word;
               ++k;
               continue;

            } else if (bracketLevel != 0 && word == "]") {
               typeArg += word;
               ++k;
               continue;

            } else if (bracketLevel != 0 && word == "[") {
               typeArg += word;
               ++k;
               continue;

            } else if (parenLevel != 0 && word == ",") {
               isStar = false;

               typeArg += word;
               ++k;
               continue;

            } else if (parenLevel != 0 && word == ")") {
               isStar = false;

               typeArg += word;
               ++k;
               continue;

            } else if (parenLevel != 0 && word == "(") {
               typeArg += word;
               ++k;
               continue;

            } else if (curlyLevel != 0 && word == "{") {
               typeArg += word;
               ++k;
               continue;

            } else if (curlyLevel != 0 && word == "}") {
               typeArg += word;
               ++k;
               continue;

            } else {

               if (k+1 < tokens.size()) {
                  nextWord = tokens[k+1];
               } else {
                  nextWord = "";
               }

               //
               if (nextWord == "*" || nextWord == "&" || nextWord == "<" || nextWord == "[" || nextWord == "(" || nextWord == "{") {
                  typeArg += word + " ";
                  ++k;
                  continue;

               }  else {
                  typeArg += word;
                  ++k;

                  if (angleLevel == 0 && bracketLevel == 0 && parenLevel == 0 && curlyLevel == 0) {
                     break;

                  } else {
                     continue;

                  }

               }
            }

            typeArg = typeArg + word + " ";
            ++k;
         }


         // C save var name
         if (typeArg.at(0) == '{') {
            // inline body, not an argument

            signature += typeArg;
            break;
         }

         word = tokens[k];

         // D
         if (found_funcPtrVar) {
            // fall thru

         } else if (word == "=" || word == ")" || word == ",")  {
            // default name
            int tmp = paramNames.size();

            QString newName = "un_named_arg" + QString::number(tmp + 1);
            paramNames.append(newName);

            signature += typeArg + " " + newName;

         } else {
            // token is a variable
            paramNames.append(word);
            ++k;

            signature += typeArg + " " + word;
         }


         // E token should be equal, paren, comma, array
         while (true) {
            word = tokens[k];

            if (word == "=") {
               signature += " " + word + " ";
               isDefaultArg = true;

            } else if (word == ")")  {
               // copy the rest as we are actually done

               while (k < tokens.size())  {
                  signature += tokens[k++];
               }

               break;

            } else if (word == ",")  {
               signature += word;
               break;

            } else if (! isDefaultArg) {
               signature += word;
               typeArg += word;

            } else   {
               signature += word;

            }

            ++k;
         }
      }
   }

   return std::make_tuple(methodName, paramNames, signature);
}

bool PepperMill::pmIs_WhiteSpace(char s)
{
   return (s == ' ' || s == '\n' || s == '\r' || s == '\t');
}

bool PepperMill::pmIs_LineBreak(char s)
{
   return (s == '\n' || s == '\r');
}

bool PepperMill::pmisToken_Method(const char **ptr)
{
   // comment, new section, closing curly -or- method
   bool retval = false;

   // never use **ptr
   const char *temp = *ptr;

  QString word;

   while (*temp) {
      char letter = *temp;

      if (letter == '}') {
         // end of class
         break;

      } else if (letter == ';')   {
         // have a method
         retval = true;
         break;

      } else if (m_sectionSet.contains(word)) {
         // new section
         if (letter == ':' || pmIs_WhiteSpace(letter) ) {
            break;
         }
      }

      // must appear after the section test
      if (pmIs_WhiteSpace(letter)) {
         ++temp;
         continue;
      }

      word += letter;
      ++temp;
   }

   return retval;
}

bool PepperMill::pmisToken_Slot(const char **ptr)
{
   bool retval   = false;
   int backTrack = 0;

   QString word;

   while (**ptr) {
      char letter = **ptr;

      if (pmIs_WhiteSpace(letter)) {
         ++backTrack;
         ++(*ptr);
         continue;

      } else if (letter == ':') {

         *ptr = *ptr - backTrack;
         break;

      } else if (word + letter == "slots") {
         retval = true;
         break;

      } else if (word + letter == "Q_SLOTS") {
         retval = true;
         break;
      }

      word += letter;

      ++backTrack;
      ++(*ptr);
   }

   return retval;
}

QString PepperMill::pmPad(const char *ptr)
{
   QString data;

   // always back up one
   --ptr;

   // a
   while (true) {
      char letter = *ptr;

      if (letter == ' ' || letter == '\t' || letter == '\n' || letter == '\r')   {
         break;
      }

      --ptr;
   }

   // b, find the padding for alignment
   while (true) {
      char letter = *ptr;

      if (letter == ' ' || letter == '\t' )   {
         // data += letter;
         data.prepend(letter);

         --ptr;
         continue;
      }

      break;
   }

   return data;
}

